/* -*- mode: c++; tab-width: 8; c-file-style: "nester" -*- */

#include "types.h"
#include "nes.h"
#include "debug.h"
#include "sdl_saving.h"
#include "bzlib.h"
#include "rc.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

extern char nes_homedir[];

static void
set_savefilename(char *str, int len, NES *emu)
{
  snprintf(str, len,
	   "%s/%s.sav",
           rc_getstr("savedir"),
	   emu->getROMname());
}


bool
sram_save (NES *emu)
{
  FILE *fp = NULL;
  char filename[512];
  const uint8 *buf;
  uint buf_len;
  uint8 compressed_buf[NES::sram_buflen];
  uint compressed_len;
  
  if (!emu->has_sram ()) return true;
  if (emu->is_invalid_sram ()) return true;
  
  buf = emu->get_SaveRAM ();
  buf_len = NES::sram_buflen;
  compressed_len = sizeof(compressed_buf);

  if (BZ2_bzBuffToBuffCompress((char *)compressed_buf,
			       (unsigned int*)&compressed_len,
			       (char *)buf, buf_len, 1, 0, 0) != BZ_OK)
  {
    return false;
  }

  set_savefilename(filename, sizeof(filename), emu);

  fp = fopen(filename, "w");
  if (!fp)
    goto error;

  if (fwrite(compressed_buf, compressed_len, 1, fp) != 1) {
    goto error;
  }
  
  fclose(fp);
  
  LOG("save done");
  return true;
  
error:
  if (fp) fclose(fp);
  return false;
}


bool
sram_load (NES *emu)
{
  FILE *fp = NULL;
  char filename[512];
  uint8 sram[NES::sram_buflen];
  int sram_len;
  int nread;
  uint8 buf[NES::sram_buflen];
  uint32 buf_len;
  uint8 compressed_buf[NES::sram_buflen];
  uint32 compressed_len;

  if (!emu->has_sram ()) return true;

  buf_len = sizeof(buf);
  compressed_len = sizeof(compressed_buf);

  set_savefilename(filename, sizeof(filename), emu);
  fp = fopen(filename, "r");
  if (!fp)
    goto error;

  compressed_len = fread(compressed_buf, 1, sizeof(compressed_buf), fp);
  if (compressed_len < 0)
    goto error;

  if (BZ2_bzBuffToBuffDecompress((char *)buf,
                                 (unsigned int*)&buf_len,
                                 (char *)compressed_buf,
                                 compressed_len, 0, 0) != BZ_OK)
  {
    goto error;
  }

  emu->set_SaveRAM (buf, buf_len);
  fclose(fp);

  LOG("load done");
  return true;

error:
  if (fp) fclose(fp);
  return false;
}


bool
disk_save (NES *emu)
{
  uint8 buf[NES::fds_save_buflen];
  uint buf_len;
  uint8 compressed_buf[NES::fds_save_buflen];
  uint compressed_len;
  char filename[512];
  FILE *fp = NULL;

  // must not save before load disk image to disk[] in mapper reset
  if(emu->GetDiskData(0) != 0x01)
    return false;

  buf_len = sizeof(buf);
  compressed_len = sizeof(compressed_buf);

  emu->make_savedata(buf);
  if (BZ2_bzBuffToBuffCompress((char *)compressed_buf,
			       (unsigned int*)&compressed_len,
			       (char *)buf, buf_len,
			       1, 0, 0) != BZ_OK)
  {
    return false;
  }

  set_savefilename(filename, sizeof(filename), emu);

  fp = fopen(filename, "w");
  if (!fp)
    goto error;

  if (fwrite(compressed_buf, compressed_len, 1, fp) != 1)
    goto error;

  fclose(fp);

  LOG ("save done");
  return true;

error:
  if (fp) fclose(fp);
  return false;
}


bool
disk_load (NES *emu)
{
  uint8 buf[NES::fds_save_buflen];
  uint32 buf_len;
  uint8 compressed_buf[NES::fds_save_buflen];
  uint32 compressed_len;
  char filename[512];
  FILE *fp = NULL;

  // must not load before load disk image to disk[] in mapper reset
  if (emu->GetDiskData(0) != 0x01)
    return false;

  buf_len = sizeof(buf);
  compressed_len = sizeof(compressed_buf);

  set_savefilename(filename, sizeof(filename), emu);
  fp = fopen(filename, "r");
  if (!fp)
    goto error;

  compressed_len = fread(compressed_buf, sizeof(compressed_buf), 1, fp);
  if (compressed_len < 0)
    goto error;

  fclose(fp);

  if (BZ2_bzBuffToBuffDecompress((char *)buf,
				 (unsigned int*)&buf_len,
				 (char *)compressed_buf,
				 compressed_len,
				 0, 0) != BZ_OK)
  {
    return false;
  }

  emu->restore_savedata (buf);
  return true;

error:
  if (fp) fclose(fp);
  return false;
}

#if 0
bool
snss_save (NES *emu)
{
  FILE *fp = NULL;
  uint8 buf[NES::snss_buflen];
  uint buf_len;
  uint8 compressed_buf[NES::snss_buflen];
  uint compressed_len;
  char filename[512];

  char templ[512];
  strcpy(templ, "/tmp/nester-XXXXXX");
  const char *tmpfilename = mktemp(templ);
  fprintf(stderr, "snss_save: %s\n", tmpfilename);

  if (!(emu->saveState(tmpfilename)))
    return false;

  fp = fopen(tmpfilename, "rb");
  if (!fp) {
    unlink(tmpfilename);
    goto error;
  }
  fseek(fp, 0, SEEK_END);
  buf_len = ftell(fp);
  fseek(fp, 0, SEEK_SET);
  if (fread(buf, buf_len, 1, fp) != 1) {
    unlink(tmpfilename);
    goto error;
  }
  fclose(fp);
  fp = NULL;

  unlink(tmpfilename);

  compressed_len = sizeof(compressed_buf);

  if (BZ2_bzBuffToBuffCompress((char *)compressed_buf,
			       (unsigned int*)&compressed_len,
			       (char *)buf,
			       buf_len,
			       1, 0, 0) != BZ_OK)
  {
    goto error;
  }

  snprintf(filename, sizeof(filename),
           "%s/%s.%d",
           rc_getstr("snapdir"),
           emu->getROMname(),
           rc_getint("saveslot"));
  fp = fopen(filename, "wb");
  if (!fp)
    goto error;
  if (fwrite(compressed_buf, compressed_len, 1, fp) != 1)
    goto error;

  return true;

error:
  if (fp) fclose(fp);
  fprintf(stderr, "snapsave failed\n");
  return false;
}
#endif

bool
snss_save (NES *emu)
{
  FILE *fp = NULL;
  uint8 buf[NES::snss_buflen];
  uint buf_len;
  char filename[512];
  int bzerror;
  BZFILE *bzfp;

  char templ[512];
  strcpy(templ, "/tmp/nester-XXXXXX");
  const char *tmpfilename = mktemp(templ);
//   fprintf(stderr, "snss_save: %s\n", tmpfilename);

  if (!(emu->saveState(tmpfilename)))
    return false;

  fp = fopen(tmpfilename, "rb");
  if (!fp) {
    unlink(tmpfilename);
    goto error;
  }
  fseek(fp, 0, SEEK_END);
  buf_len = ftell(fp);
  fseek(fp, 0, SEEK_SET);
  if (fread(buf, buf_len, 1, fp) != 1) {
    unlink(tmpfilename);
    goto error;
  }
  fclose(fp);
  fp = NULL;
  unlink(tmpfilename);

  snprintf(filename, sizeof(filename),
           "%s/%s.%d",
           rc_getstr("snapdir"),
           emu->getROMname(),
           rc_getint("saveslot"));
  fp = fopen(filename, "wb");
  if (!fp)
    goto error;

  bzfp = BZ2_bzWriteOpen(&bzerror, fp, 1, 0, 0);
  BZ2_bzWrite(&bzerror, bzfp, buf, buf_len);
  BZ2_bzWriteClose(&bzerror, bzfp, 0, 0, 0);

  return true;

error:
  if (fp) fclose(fp);
  fprintf(stderr, "snapsave failed\n");
  return false;
}

bool
snss_load (NES *emu)
{
  FILE *fp = NULL;
  uint8 buf[NES::snss_buflen];
  size_t buf_len;
  char filename[512];
  int bzerror;
  BZFILE *bzfp;

  char templ[512];
  strcpy(templ, "/tmp/nester-XXXXXX");
  const char *tmpfilename = mktemp(templ);
//   fprintf(stderr, "snss_load: %s\n", tmpfilename);

  buf_len = sizeof(buf);

  snprintf(filename, sizeof(filename),
           "%s/%s.%d",
           rc_getstr("snapdir"),
           emu->getROMname(),
           rc_getint("saveslot"));

  fp = fopen(filename, "rb");
  if (!fp)
    goto error;

  bzfp = BZ2_bzReadOpen(&bzerror, fp, 0, 0, 0, 0);
  if (!bzfp)
    goto error;

  buf_len = BZ2_bzRead(&bzerror, bzfp, buf, buf_len);
  BZ2_bzReadClose(&bzerror, bzfp);
  fclose(fp);

  fp = fopen(tmpfilename, "w");
  if (!fp)
    goto error;
  if (fwrite(buf, buf_len, 1, fp) != 1) {
    unlink(tmpfilename);
    goto error;
  }
  fclose(fp);
  fp = NULL;

  if (!emu->loadState(tmpfilename)) {
    unlink(tmpfilename);
    return false;
  }
  unlink(tmpfilename);

  return true;

error:
  if (fp) fclose(fp);
  fprintf(stderr, "snapload failed\n");
  return false;
}

#if 0
bool
snss_load (NES *emu)
{
  uint8 compressed_buf[NES::snss_buflen];
  uint32 compressed_len;
  uint8 buf[NES::snss_buflen];
  size_t buf_len;
  char filename[512];
  FILE *fp = NULL;
  char templ[512];
  strcpy(templ, "nester-XXXXXX");
  const char *tmpfilename = mktemp(templ);
  fprintf(stderr, "snss_load: %s\n", tmpfilename);

  compressed_len = sizeof(compressed_buf);
  buf_len = sizeof(buf);

  snprintf(filename, sizeof(filename),
           "%s/%s.%d",
           rc_getstr("snapdir"),
           emu->getROMname(),
           rc_getint("saveslot"));

  fp = fopen(filename, "rb");
  if (!fp)
    goto error;

  if ((compressed_len = fread(compressed_buf, sizeof(compressed_buf), 1, fp))
      < 0)
    goto error;

  fclose(fp);

  if (BZ2_bzBuffToBuffDecompress((char *)buf, (unsigned int*)&buf_len,
				 (char *)compressed_buf, compressed_len,
				 0, 0) != BZ_OK)
    goto error;

  fp = fopen(tmpfilename, "wb");
  if (!fp)
    goto error;
  if (fwrite(buf, buf_len, 1, fp) != 1) {
    unlink(tmpfilename);
    goto error;
  }
  fclose(fp);
  fp = NULL;

  if (!emu->loadState(tmpfilename)) {
    unlink(tmpfilename);
    return false;
  }
  unlink(tmpfilename);

  return true;

error:
  if (fp) fclose(fp);
  fprintf(stderr, "snapload failed\n");
  return false;
}
#endif
